import json
import os

from coworks import Blueprint
from pygsheets import Worksheet, DataRange
from pygsheets import authorize, WorksheetNotFound


class GoogleSheets(Blueprint):

    def __init__(self, client_id_var_name, private_key_var_name, private_key_id_var_name, **kwargs):
        super().__init__(**kwargs)
        self.client_id_var_name = client_id_var_name
        self.private_key_var_name = private_key_var_name
        self.private_key_id_var_name = private_key_id_var_name
        self.google_client = None

        @self.before_first_activation
        def load_crendentials(event, context):
            client_id = os.getenv(self.client_id_var_name)
            if not client_id:
                raise EnvironmentError(f'{self.client_id_var_name} not defined in environment.')
            private_key = os.getenv(self.private_key_var_name)
            if not private_key:
                raise EnvironmentError(f'{self.private_key_var_name} not defined in environment.')
            private_key_id = os.getenv(self.private_key_id_var_name)
            if not private_key_id:
                raise EnvironmentError(f'{self.private_key_id_var_name} not defined in environment.')

            self.CREDENTIALS = {
                "type": "service_account",
                "project_id": "glassy-bonsai-277705",
                "private_key_id": private_key_id,
                "private_key": private_key,
                "client_email": "google-sheet-micro-service@glassy-bonsai-277705.iam.gserviceaccount.com",
                "client_id": client_id,
                "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                "token_uri": "https://oauth2.googleapis.com/token",
                "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/google-sheet-micro-service%40glassy-bonsai-277705.iam.gserviceaccount.com"
            }

            os.environ['SERVICE_ACCOUNT_CREDENTIALS'] = json.dumps(self.CREDENTIALS)
            self.google_client = authorize(service_account_env_var="SERVICE_ACCOUNT_CREDENTIALS")

    def get_worksheets(self, key: str):
        """Get list of worksheet index and titles.
        :param key:         Document google key.
        """
        spreadsheet = self.google_client.open_by_key(key=key)
        return [{'index': w.index, 'title': w.title} for w in spreadsheet.worksheets()]

    def get_worksheet_all(self, key: str, worksheet: str, property='index', **kwargs):
        """Get data from worksheet.
        :param key:         Document google key.
        :param worksheet:   Worksheet property.
        :param property:    May be index, title or id.
        """
        return self._get_worksheet(key, worksheet, property=property).get_all_records(**kwargs)

    def get_worksheet_value(self, key: str, worksheet: str, addr, property='index'):
        """Get data from worksheet.
        :param key:         Document google key.
        :param worksheet:   Worksheet property.
        :param property:    May be index, title or id.
        :param start:       Top left cell address. Can be unbounded.
        :param end:         Bottom right cell address.
        """
        return self._get_worksheet(key, worksheet, property=property).get_value(addr)

    def get_worksheet_values(self, key: str, worksheet: str, property='index', start=(1, 1), end=(1, None), **kwargs):
        """Get data from worksheet.
        :param key:         Document google key.
        :param worksheet:   Worksheet property.
        :param property:    May be index, title or id.
        :param start:       Top left cell address. Can be unbounded.
        :param end:         Bottom right cell address.
        """
        return self._get_worksheet(key, worksheet, property=property).get_values(start, end, **kwargs)

    def post_worksheet(self, key: str, index=0, title='', **kwargs):
        """Add a new worksheet.
        :param key:     Document google key.
        :param index:   Tab index of the worksheet.
        :param title:   Worksheet's title.
        """
        spreadsheet = self.google_client.open_by_key(key=key)
        spreadsheet.add_worksheet(title, index=index, **kwargs)

    def post_worksheet_content(self, key: str, worksheet: str, property='index', values=None, start=(1, 1), **kwargs):
        """Write data in worksheet.
        """
        worksheet = self._get_worksheet(key, worksheet, property=property)
        worksheet.update_values(start, values, **kwargs)

    def _get_worksheet(self, key, worksheet, *, property) -> Worksheet:
        spreadsheet = self.google_client.open_by_key(key=key)
        worksheet = spreadsheet.worksheet(property=property, value=worksheet)
        return worksheet



import inspect
import pathlib
import sys

import click
from jinja2 import Environment, PackageLoader, select_autoescape, TemplateNotFound

from .command import CwsCommand
from .error import CwsCommandError


class CwsWriterError(CwsCommandError):
    ...


class CwsWriter(CwsCommand):

    @property
    def options(self):
        return [
            *super().options,
            click.option('-o', '--output'),
            click.option('--debug/--no-debug', default=False, help='Print debug logs to stderr.')
        ]

    @classmethod
    def multi_execute(cls, project_dir, workspace, execution_list):
        for command, options in execution_list:
            command._export_header(**options)
            command._export_content(**options)

    def __init__(self, app=None, name='export'):
        super().__init__(app, name=name)

    def _export_header(self, **options):
        print("// Do NOT edit this file as it is auto-generated by cws\n", file=self.output, flush=True)

    def _export_content(self, **options):
        """ Main export function.
        :param options: Command options.

        Abstract method which must be redefined in any subclass. The content should be written in self.output.
        """
        if options['debug']:
            print(f"Print content in {self.output.name}")


class CwsTemplateWriter(CwsWriter):
    """Writer with  jinja templating."""

    def __init__(self, app=None, *, name='export', data=None, template=None, env=None):
        super().__init__(app, name=name)
        self.data = data or {}
        self.template_filenames = template or []
        self.env = env or Environment(
            loader=PackageLoader(sys.modules[__name__].__name__),
            autoescape=select_autoescape(['html', 'xml']))

    @property
    def options(self):
        return [
            *super().options,
            click.option('--template', '-t', multiple=True),
        ]

    def _export_content(self, *, project_dir, module, service, workspace, **options):
        super()._export_content(**options)

        module_path = module.split('.')
        template_filenames = options.get('template') or self.template_filenames

        # Get parameters for execution
        config = self.app.get_config(workspace)
        if config is None:
            raise CwsCommandError("A workspace is mandatory in the python configuration for deploying.\n")

        environment_variable_files = [p.as_posix() for p in
                                      config.existing_environment_variables_files(project_dir)]
        data = {
            'writer': self,
            'project_dir': project_dir,
            'source_file': pathlib.PurePath(project_dir, *module_path),
            'module': module,
            'module_path': pathlib.PurePath(*module_path),
            'module_dir': pathlib.PurePath(*module_path[:-1]),
            'module_file': module_path[-1],
            'handler': service,
            'app': self.app,
            'ms_name': self.app.name,
            'workspace': workspace,
            'app_config': config,
            'environment_variables': config.environment_variables,
            'environment_variable_files': environment_variable_files,
            'account_number': options.get('account_number'),
            'description': inspect.getdoc(self.app) or "",
            **options
        }
        data.update(self.data)
        try:
            for template_filename in template_filenames:
                template = self.env.get_template(template_filename)
                print(template.render(**data), file=self.output, flush=True)
        except TemplateNotFound as e:
            raise CwsWriterError(f"Cannot find template {str(e)}")
        except Exception as e:
            raise CwsWriterError(e)



import cgi
import inspect
import io
import json
import traceback
import urllib
from functools import update_wrapper, partial

from aws_xray_sdk.core import xray_recorder
from botocore.exceptions import BotoCoreError
from chalice import AuthResponse, Response, ChaliceViewError
from requests_toolbelt.multipart import MultipartDecoder

from .aws import AwsS3Session
from .error import CwsError
from .utils import class_auth_methods, class_cws_methods, make_absolute, path_join


class CoworksMixin:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.debug = kwargs.pop('debug', False)

        # A list of functions that will be called at the first activation.
        # To register a function, use the :meth:`before_first_request` decorator.
        self.before_first_activation_funcs = []

        # A list of functions that will be called at the beginning of each activation.
        # To register a function, use the :meth:`before_activation` decorator.
        self.before_activation_funcs = []

        # A list of functions that will be called after each activation.
        # To register a function, use the :meth:`after_activation` decorator.
        self.after_activation_funcs = []

        # A list of functions that will be called in case of exception.
        # To register a function, use the :meth:`handle_exception` decorator.
        self.handle_exception_funcs = []

        self.aws_s3_sfn_data_session = AwsS3Session(env_var_access_key="AWS_RUN_ACCESS_KEY_ID",
                                                    env_var_secret_key="AWS_RUN_SECRET_KEY",
                                                    env_var_region="AWS_RUN_REGION")
        self.aws_s3_form_data_session = AwsS3Session(env_var_access_key="AWS_FORM_DATA_ACCESS_KEY_ID",
                                                     env_var_secret_key="AWS_FORM_DATA_SECRET_KEY",
                                                     env_var_region="AWS_FORM_DATA_REGION")

    def before_first_activation(self, f):
        """Registers a function to be run before the first activation of the microservice.

        May be used as a decorator.

        The function will be called with event and context positional arguments and its return value is ignored.
        """

        self.before_first_activation_funcs.append(f)
        return f

    def before_activation(self, f):
        """Registers a function to called before each activation of the microservice.
        :param f:  Function added to the list.
        :return: None.

        May be used as a decorator.

        The function will be called with event and context positional arguments and its return value is ignored.
        """

        self.before_activation_funcs.append(f)
        return f

    def after_activation(self, f):
        """Registers a function to be called after each activation of the microservice.

        May be used as a decorator.

        The function will be called with the response and its return value is the final response.
        """

        self.after_activation_funcs.append(f)
        return f

    def handle_exception(self, f):
        """Registers a function to be called in case of exception.

        May be used as a decorator.

        The function will be called with the exception and eventually a response to return.
        """

        self.handle_exception_funcs.append(f)
        return f

    def _init_routes(self, app, *, url_prefix='', hide_routes=False):
        """ Creates all routes for a microservice.
        :param hide_routes list of routes to be hidden.
        """

        # Global authorization function may be redefined
        auth_fun = app.config.auth if app.config.auth else class_auth_methods(app)
        auth = self._create_auth_proxy(app, auth_fun) if auth_fun else None

        # Adds entrypoints
        methods = class_cws_methods(self)
        for fun in methods:
            if hide_routes is True or getattr(fun, '__CWS_HIDDEN', False):
                continue

            method = getattr(fun, '__CWS_METHOD')
            path = getattr(fun, '__CWS_PATH')
            entry_path = route = path_join(url_prefix, path)

            # Get parameters
            args = inspect.getfullargspec(fun).args[1:]
            defaults = inspect.getfullargspec(fun).defaults
            varkw = inspect.getfullargspec(fun).varkw
            if defaults:
                len_defaults = len(defaults)
                for index, arg in enumerate(args[:-len_defaults]):
                    entry_path = path_join(entry_path, f"/{{{arg}}}")
                    route = path_join(route, f"/{{_{index}}}")
                kwarg_keys = args[-len_defaults:]
            else:
                for index, arg in enumerate(args):
                    entry_path = path_join(entry_path, f"/{{{arg}}}")
                    route = path_join(route, f"/{{_{index}}}")
                kwarg_keys = {}

            proxy = self._create_rest_proxy(fun, kwarg_keys, args, varkw)

            # Creates the entry
            if hide_routes is False or (type(hide_routes) is list and entry_path not in hide_routes):
                if app.entries is None:
                    app.entries = {}
                if entry_path not in app.entries:
                    app.entries[entry_path] = {}
                if method in app.entries[entry_path]:
                    raise CwsError(f"The method {method} is already defined for the route {make_absolute(entry_path)}")
                app.add_entry(entry_path, method, auth, fun)
                app.route(f"{make_absolute(route)}", methods=[method], authorizer=auth, cors=app.config.cors,
                          content_types=list(app.config.content_type))(proxy)

    def _create_auth_proxy(self, app, auth_method):

        def proxy(auth_activation):
            subsegment = xray_recorder.current_subsegment()
            try:
                auth = auth_method(auth_activation)
                if subsegment:
                    subsegment.put_metadata('result', auth)
            except Exception as e:
                app.log.info(f"Exception : {str(e)}")
                traceback.print_exc()
                if subsegment:
                    subsegment.add_exception(e, traceback.extract_stack())
                return AuthResponse(routes=[], principal_id='user')

            if type(auth) is bool:
                if auth:
                    return AuthResponse(routes=['*'], principal_id='user')
                return AuthResponse(routes=[], principal_id='user')
            elif type(auth) is list:
                return AuthResponse(routes=auth, principal_id='user')
            return auth

        proxy = update_wrapper(proxy, auth_method)
        proxy.__name__ = 'app'
        return self.authorizer(name='auth')(proxy)

    def _create_rest_proxy(self, func, kwarg_keys, args, varkw):
        original_app_class = self.__class__

        def proxy(**kws):
            try:
                # Renames positional parameters (index added in label)
                kwargs = {}
                for kw, value in kws.items():
                    param = args[int(kw[1:])]
                    kwargs[param] = value

                # Adds kwargs parameters
                def check_param_expected_in_lambda(param_name):
                    """Alerts when more parameters than expected are defined in request."""
                    if param_name not in kwarg_keys and varkw is None:
                        err_msg = f"TypeError: got an unexpected keyword argument '{param_name}'"
                        return Response(body=err_msg, status_code=400)

                def add_param(param_name, param_value):
                    check_param_expected_in_lambda(param_name)
                    if param_name in params:
                        if isinstance(params[param_name], list):
                            params[param_name].append(param_value)
                        else:
                            params[param_name] = [params[param_name], param_value]
                    else:
                        params[param_name] = param_value

                req = self.current_request
                if kwarg_keys or varkw:
                    params = {}
                    if req.raw_body:  # POST request
                        try:
                            content_type = req.headers['content-type']
                            if content_type.startswith('multipart/form-data'):
                                try:
                                    multipart_decoder = MultipartDecoder(req.raw_body, content_type)
                                    for part in multipart_decoder.parts:
                                        name, content = self._get_multipart_content(part)
                                        add_param(name, content)
                                except Exception as e:
                                    return Response(body=str(e), status_code=400)
                                kwargs = dict(**kwargs, **params)
                            elif content_type.startswith('application/x-www-form-urlencoded'):
                                params = urllib.parse.parse_qs(req.raw_body.decode("utf-8"))
                                kwargs = dict(**kwargs, **params)
                            elif content_type.startswith('application/json'):
                                if hasattr(req.json_body, 'items'):
                                    params = {}
                                    for k, v in req.json_body.items():
                                        add_param(k, v)
                                    kwargs = dict(**kwargs, **params)
                                else:
                                    kwargs[kwarg_keys[0]] = req.json_body
                            elif content_type.startswith('text/plain'):
                                kwargs[kwarg_keys[0]] = req.json_body
                            else:
                                err = f"Cannot manage content type {content_type} for {self}"
                                return Response(body=err, status_code=400)
                        except Exception as e:
                            return Response(body=str(e), status_code=400)

                    else:  # GET request

                        # adds parameters from qurey parameters
                        for k in req.query_params or []:
                            value = req.query_params.getlist(k)
                            add_param(k, value if len(value) > 1 else value[0])
                        kwargs = dict(**kwargs, **params)
                else:
                    if not args and (req.raw_body or req.query_params):
                        err = f"TypeError: got an unexpected arguments (body: {req.raw_body}, query: {req.query_params}"
                        return Response(body=err, status_code=400)

                # chalice is changing class for local server for threading reason (why not mixin..?)
                self_class = self.__class__
                if self_class != original_app_class:
                    self.__class__ = original_app_class

                resp = func(self, **kwargs)
                self.__class__ = self_class
                return self._convert_response(resp)
            except ChaliceViewError:
                raise
            except TypeError as e:
                return Response(body=str(e), status_code=400)
            except Exception:
                subsegment = xray_recorder.current_subsegment()
                if subsegment:
                    subsegment.add_error_flag()
                raise

        proxy = update_wrapper(proxy, func)
        proxy.__cws_func__ = update_wrapper(partial(func, self), func)
        return proxy

    def _get_multipart_content(self, part):
        headers = {k.decode('utf-8'): cgi.parse_header(v.decode('utf-8')) for k, v in part.headers.items()}
        content = part.content
        _, content_disposition_params = headers['Content-Disposition']
        part_content_type, _ = headers.get('Content-Type', (None, None))
        name = content_disposition_params['name']

        # content in a text or json value
        if 'filename' not in content_disposition_params:
            if part_content_type == 'application/json':
                return name, self._get_data_on_s3(json.loads(content.decode('utf-8')))
            return name, self._get_data_on_s3(content.decode('utf-8'))

        # content in a file (s3 or plain text)
        if part_content_type == 'text/s3':
            pathes = content.decode('utf-8').split('/', 1)
            try:
                s3_object = self.aws_s3_form_data_session.client.get_object(Bucket=pathes[0], Key=pathes[1])
            except BotoCoreError:
                return CwsError(f"Bucket={pathes[0]} Key={pathes[1]} not found on s3")
            file = io.BytesIO(s3_object['Body'].read())
            mime_type = s3_object['ContentType']
        else:
            file = io.BytesIO(content)
            mime_type = part_content_type
        file.name = content_disposition_params['filename']

        return name, FileParam(file, mime_type)

    def _set_multipart_content(self, form_data):
        def encode_part(_part):
            if type(_part) is str:
                return None, _part, 'text/plain'

            if 'mime_type' in _part:
                mime_type = _part.get('mime_type')
            elif 'json' in _part:
                _part['content'] = _part.get('json')
                mime_type = 'application/json'
            elif 's3' in _part:
                path = _part.get('s3')
                _part['filename'] = path.split('/')[-1]
                _part['path'] = path
                mime_type = 'text/s3'
            else:
                mime_type = 'text/plain'

            filename = _part.get('filename')
            if mime_type == 'text/plain':
                content = _part.get('content')
                return filename, content, mime_type
            elif mime_type == 'application/json':
                content = _part.get('content')
                return filename, json.dumps(content), mime_type
            elif mime_type == 'text/s3':
                path = _part.get('path')
                return filename, path, mime_type
            else:
                return Response(body=f"Undefined mime type {mime_type}", status_code=400)

        parts = []
        for name, part in form_data.items():
            if type(part) is list:
                parts.extend([(name, encode_part(p)) for p in part])
            else:
                parts.append((name, encode_part(part)))
        return parts

    def _set_data_on_s3(self, data):
        """Saves value on S3 temporary file if content is too big."""

        def set_on_s3(value):
            s3_client = self.aws_s3_sfn_data_session.client
            context = self.lambda_context
            key = f"tmp/{context.aws_request_id}"
            tags = f"Name={context.function_name}"
            s3_client.put_object(Bucket="coworks-microservice", Key=key, Body=value, Tagging=tags)
            return f"$${key}$$"

        if type(data) == str:
            return set_on_s3(data) if len(data) > 1000 else data
        if type(data) == list:
            return [self._set_data_on_s3(v) for v in data]
        if type(data) == dict:
            for k, v in data.items():
                data[k] = self._set_data_on_s3(v)
        return data

    def _get_data_on_s3(self, data):
        """Retrieves value from S3 temporary file (content too big)."""

        def get_on_s3(value):
            s3_client = self.aws_s3_sfn_data_session.client
            s3_object = s3_client.get_object(Bucket="coworks-microservice", Key=value[2:-2])
            return s3_object['Body'].read().decode("utf-8")

        if type(data) == str:
            return get_on_s3(data) if data.startswith('$$') and data.endswith('$$') else data
        if type(data) == list:
            return [self._get_data_on_s3(v) for v in data]
        if type(data) == dict:
            for k, v in data.items():
                data[k] = self._get_data_on_s3(v)
        return data

    @staticmethod
    def _convert_response(resp):
        if type(resp) is tuple:
            if len(resp) == 2:
                if type(resp[1]) is int:
                    return Response(body=resp[0], status_code=resp[1])
                elif type(resp[1]) is dict:
                    return Response(body=resp[0], status_code=200, headers=resp[1])
                else:
                    return Response(body="Internal error (wrong result type)", status_code=500)
            else:
                return Response(body=resp[0], status_code=resp[1], headers=resp[2])

        return resp


class FileParam:

    def __init__(self, file, mime_type):
        self.file = file
        self.mime_type = mime_type

    def __repr__(self):
        if self.mime_type:
            return f'FileParam({self.file.name}, {self.mime_type})'
        return f'FileParam({self.file.name})'



import os

import sqlalchemy
from sqlalchemy import create_engine, text, MetaData, Table
from sqlalchemy.orm import sessionmaker

from ..coworks import TechMicroService


class SqlMicroService(TechMicroService):

    def __init__(self, *, dialect=None, host=None, port=None, dbname=None, user=None, **kwargs):
        super().__init__(**kwargs)
        self.dialect = dialect
        self.host = host
        self.port = port
        self.dbname = dbname
        self.user = user
        self.password = None
        self.session = None

        self.__engine = None

        @self.before_first_activation
        def check_env_vars(event, context):
            if self.dialect is None:
                self.dialect = os.getenv(self.dialect_env_var_name)
                if not self.dialect:
                    raise EnvironmentError(f"{self.dialect_env_var_name} not defined in environment.")
            if self.host is None:
                self.host = os.getenv(self.host_env_var_name)
                if not self.host:
                    raise EnvironmentError(f"{self.host_env_var_name} not defined in environment.")
            if self.port is None:
                self.port = os.getenv(self.port_env_var_name)
                if self.port is None:
                    if self.dialect.startswith('mysql'):
                        self.port = 3306
                    elif self.dialect == 'postgres':
                        self.port = 5432
            if self.dbname is None:
                self.dbname = os.getenv(self.dbname_env_var_name)
                if not self.dbname:
                    raise EnvironmentError(f"{self.dbname_env_var_name} not defined in environment.")
            if self.user is None:
                self.user = os.getenv(self.user_env_var_name)
                if not self.user:
                    raise EnvironmentError(f"{self.user_env_var_name} not defined in environment.")
            self.password = os.getenv(self.passwd_env_var_name, '')

        @self.before_activation
        def set_session(event, context):
            self.session = sessionmaker(bind=self.engine)()

    @property
    def engine(self):
        if not self.__engine:
            self.__engine = create_engine(
                f'{self.dialect}://{self.user}:{self.password}@{self.host}:{self.port}/{self.dbname}', echo=False
            )
        return self.__engine

    def get_version(self):
        """Returns SQLAlchemy version."""
        return sqlalchemy.__version__, {'content-type': 'text/plain'}

    def get_fetch(self, query: str = None, **kwargs):
        conn = self.engine.connect()
        rows = conn.execute(text(query), **kwargs).fetchall()
        return [dict(row) for row in rows]

    def reflect_tables(self, schema, tables):
        metadata = MetaData(bind=self.engine)
        return list(
            map(lambda table: Table(table, metadata, autoload=True, autoload_with=self.engine, schema=schema), tables))

    @property
    def dialect_env_var_name(self):
        return 'DIALECT'

    @property
    def host_env_var_name(self):
        return 'HOST'

    @property
    def port_env_var_name(self):
        return 'PORT'

    @property
    def dbname_env_var_name(self):
        return 'DBNAME'

    @property
    def user_env_var_name(self):
        return 'USER'

    @property
    def passwd_env_var_name(self):
        return 'PASSWD'



import importlib
import inspect
import json
import os
import platform
import sys
from functools import partial

HTTP_METHODS = ['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'OPTIONS']


def jsonify(result, pretty=False, indent=None, separators=None):
    if pretty:
        indent = indent or 4
        separators = separators or (",", ": ")
    else:
        separators = separators or (",", ":")

    return json.dumps(result, indent=indent, separators=separators)


def import_attr(module, attr, cwd='.'):
    if cwd not in sys.path:
        sys.path.insert(0, cwd)
    app_module = importlib.import_module(module)
    if "PYTEST_CURRENT_TEST" in os.environ:
        # needed as Chalice local server change class
        app_module = importlib.reload(app_module)
    return getattr(app_module, attr)


def class_auth_methods(obj):
    """Returns the auth method from the class if exists."""
    methods = inspect.getmembers(obj.__class__, lambda x: inspect.isfunction(x))

    for name, func in methods:
        if name == 'auth':
            function_is_static = isinstance(inspect.getattr_static(obj.__class__, func.__name__), staticmethod)
            if function_is_static:
                return func
            return partial(func, obj)
    return None


def class_cws_methods(obj):
    """Returns the list of methods from the class."""
    methods = inspect.getmembers(obj.__class__, lambda x: inspect.isfunction(x))

    return [fun for _, fun in methods if hasattr(fun, '__CWS_METHOD')]


def class_attribute(obj, name: str = None, defaut=None):
    """Returns the list of attributes from the class or the attribute if name parameter is defined
    or default value if not found."""
    attributes = inspect.getmembers(obj.__class__, lambda x: not inspect.isroutine(x))

    if not name:
        return attributes

    filtered = [a[1] for a in attributes if a[0] == name]
    return filtered[0] if filtered else defaut


def path_join(*args):
    """ Joins given arguments into an entry route.
    Trailing but not leading slashes are stripped for each argument.
    """

    reduced = [x.lstrip('/').rstrip('/') for x in args if x]
    return '/'.join([x for x in reduced if x])


def make_absolute(route):
    return '/' + path_join(route)


def trim_underscores(name):
    while name.startswith('_'):
        name = name[1:]
    while name.endswith('_'):
        name = name[:-1]
    return name


def as_list(var):
    if var is None:
        return []
    if type(var) is list:
        return var
    return [var]


def get_system_info():
    python_info = f"python {sys.version_info[0]}.{sys.version_info[1]}.{sys.version_info[2]}"
    platform_system = platform.system().lower()
    platform_release = platform.release()
    platform_info = f"{platform_system} {platform_release}"
    return f"{python_info}, {platform_info}"



from chalice import ChaliceViewError


class CwsError(Exception):
    def __init__(self, message):
        self.msg = message


class BadRequestError(ChaliceViewError):
    STATUS_CODE = 400


class UnauthorizedError(ChaliceViewError):
    STATUS_CODE = 401


class ForbiddenError(ChaliceViewError):
    STATUS_CODE = 403


class NotFoundError(ChaliceViewError):
    STATUS_CODE = 404


class MethodNotAllowedError(ChaliceViewError):
    STATUS_CODE = 405


class RequestTimeoutError(ChaliceViewError):
    STATUS_CODE = 408


class ConflictError(ChaliceViewError):
    STATUS_CODE = 409


class UnprocessableEntityError(ChaliceViewError):
    STATUS_CODE = 422


class TooManyRequestsError(ChaliceViewError):
    STATUS_CODE = 429


class InternalServerError(ChaliceViewError):
    STATUS_CODE = 500



import os
import smtplib
from email.utils import formataddr
from email.message import EmailMessage

import requests
from aws_xray_sdk.core import xray_recorder

from coworks import Blueprint, FileParam, entry, MicroServiceProxy


class Mail(Blueprint):
    """Mail blueprint.
    Environment variables needed:
    - env_server_var_name: Variable name for the SMTP server.
    - env_login_var_name: Variable name for the login.
    - env_passwd_var_name: Variable name for the password.
    """

    def __init__(self, env_server_var_name, env_login_var_name, env_passwd_var_name, **kwargs):
        super().__init__(**kwargs)
        self.env_server_var_name = env_server_var_name
        self.env_login_var_name = env_login_var_name
        self.env_passwd_var_name = env_passwd_var_name
        self.smtp_server = self.smtp_login = self.smtp_passwd = None

        @self.before_first_activation
        def check_env_vars(event, context):
            self.smtp_server = os.getenv(self.env_server_var_name)
            if not self.smtp_server:
                raise EnvironmentError(f'{self.env_server_var_name} not defined in environment.')
            self.smtp_login = os.getenv(self.env_login_var_name)
            if not self.smtp_login:
                raise EnvironmentError(f'{self.env_login_var_name} not defined in environment.')
            self.smtp_passwd = os.getenv(env_passwd_var_name)
            if not self.smtp_passwd:
                raise EnvironmentError(f'{self.env_passwd_var_name} not defined in environment.')

    @entry
    @xray_recorder.capture()
    def post_send(self, subject="", from_addr: str = None, from_name: str = '', to_addrs: [str] = None,
                  cc_addrs: [str] = None, bcc_addrs: [str] = None, body="",
                  attachments: [FileParam] = None, attachment_urls: dict = None, subtype="plain", starttls=True):
        """ Send mail.
        To send attachments, add files in the body of the request as multipart/form-data.
        """

        from_addr = from_addr or os.getenv('from_addr')
        if not from_addr:
            return "From address not defined (from_addr:str)", 400
        to_addrs = to_addrs or os.getenv('to_addrs')
        if not to_addrs:
            return "To addresses not defined (to_addrs:[str])", 400

        # Creates email
        try:
            from_ = formataddr((from_name if from_name else False, from_addr))
            msg = EmailMessage()
            msg['Subject'] = subject
            msg['From'] = from_
            msg['To'] = to_addrs if isinstance(to_addrs, str) else ', '.join(to_addrs)
            if cc_addrs:
                msg['Cc'] = cc_addrs if isinstance(cc_addrs, str) else ', '.join(cc_addrs)
            if bcc_addrs:
                msg['Bcc'] = bcc_addrs if isinstance(bcc_addrs, str) else ', '.join(bcc_addrs)
            msg.set_content(body, subtype=subtype)

            if attachments:
                if not isinstance(attachments, list):
                    attachments = [attachments]
                for attachment in attachments:
                    if not attachment.mime_type:
                        return f"Mime type of the attachment {attachment.file.name} is not defined.", 400
                    maintype, subtype = attachment.mime_type.split("/")
                    msg.add_attachment(attachment.file.read(), maintype=maintype, subtype=subtype,
                                       filename=attachment.file.name)

            if attachment_urls:
                for attachment_name, attachment_url in attachment_urls.items():
                    response = requests.get(attachment_url)
                    if response.status_code == 200:
                        attachment = response.content
                        maintype, subtype = response.headers['Content-Type'].split('/')
                        msg.add_attachment(attachment, maintype=maintype, subtype=subtype,
                                           filename=attachment_name)
                    else:
                        return f"Failed to download attachment, error {response.status_code}.", 400

        except Exception as e:
            return f"Cannot create email message (Error: {str(e)}).", 400

        # Send email
        try:
            with smtplib.SMTP(self.smtp_server) as server:
                if starttls:
                    server.starttls()
                server.login(self.smtp_login, self.smtp_passwd)
                server.send_message(msg)

            return f"Mail sent to {msg['To']}"
        except smtplib.SMTPAuthenticationError:
            return "Wrong username/password : cannot connect.", 400
        except Exception as e:
            return f"Cannot send email message (Error: {str(e)}).", 400


class MailProxy(MicroServiceProxy):

    def send(self, path, data, attachments=None, sync=False):
        headers = {'content-type': 'multipart/form-data'} if attachments else {}
        return super().post(f'{self.url}/send', data=data, attachments=attachments, sync=sync)



import click

from .command import CwsCommand, CwsCommandError
from .. import aws


class CwsInspector(CwsCommand):
    """Command to get information on a microservice."""

    @classmethod
    def multi_execute(cls, project_dir, workspace, execution_list):
        for command, command_options in execution_list:
            command.print_module_info(**command_options)
            if command_options['env']:
                command.print_env_vars(**command_options)

    def __init__(self, app=None, name='inspect'):
        super().__init__(app, name=name)

    @property
    def options(self):
        return [
            *super().options,
            click.option('--env', '-e', is_flag=True, help="Show environment variables."),
        ]

    def print_module_info(self, *, module, service, **options):
        print(f"Microservice {service} defined in module {module}")

    def print_env_vars(self, *, project_dir, workspace, service, **options):
        for config in self.app.configs:
            if config.is_valid_for(workspace):
                print(f"Environment vars for {service} in workspace {config.workspace}:")
                files = config.existing_environment_variables_files(project_dir)
                for file in files:
                    with file.open() as f:
                        print(f.read())


class CwsAWSInspector(CwsCommand):
    """Command to get information on the project's microservices and teir deployment."""

    @classmethod
    def multi_execute(cls, project_dir, workspace, execution_list):
        for command, command_options in execution_list:
            profile_name = command_options['profile_name']
            aws_api_session = aws.ApiGatewaySession(profile_name=profile_name)
            paginator = aws_api_session.client.get_paginator('get_rest_apis')
            apis = command.get_apis(command.app.name, paginator)
            for api in apis:
                print(command.get_result(api))

    def __init__(self, app=None, name='deployed'):
        super().__init__(app, name=name)

    @property
    def options(self):
        return [
            *super().options,
            click.option('--profile-name', '-p', required=True, help="AWS profile for APIGateway access."),
            click.option('--env', '-e', is_flag=True, help="Show environment variables."),
        ]

    def get_apis(self, service_name, paginator, page_size=10):
        rest_apis = paginator.paginate(PaginationConfig={'PageSize': page_size})
        return rest_apis.search(f"items[?starts_with(name, '{service_name}')]")

    def get_result(self, api):
        return {
            'name': api['name'],
            'id': api['id'],
        }



import inspect
import sys
from inspect import Parameter

from jinja2 import Environment, PackageLoader, select_autoescape

from coworks import Blueprint, entry, jsonify
from coworks.utils import make_absolute


class Admin(Blueprint):

    @entry
    def get_route(self, pretty=False):
        """Returns the list of entrypoints with signature."""
        routes = {}
        for path, entrypoint in self._current_app.entries.items():
            route = {}
            for http_method, route_entry in entrypoint.items():
                function_called = route_entry.fun
                doc = inspect.getdoc(function_called)
                route[http_method] = {
                    'doc': doc if doc else '',
                    'signature': get_signature(function_called)
                }
            routes[make_absolute(path)] = route

        return jsonify(routes, pretty)

    @entry
    def get_context(self):
        """Returns the calling context."""
        return self.current_request.to_dict()

    @entry
    def get_proxy(self):
        """Returns the calling context."""
        env = Environment(
            loader=PackageLoader(sys.modules[__name__].__name__),
            autoescape=select_autoescape(['html', 'xml']))
        env.filters["signature"] = inspect.signature
        env.filters["positional_params"] = positional_params
        env.filters["keyword_params"] = keyword_params

        data = {
            'name': self._current_app.name,
            'entries': self._current_app.entries,
        }
        template = env.get_template("proxy.j2")
        return template.render(**data)


def get_signature(func):
    sig = ""
    params = inspect.signature(func).parameters
    for i, (k, p) in enumerate(params.items()):
        if i == 0:
            continue
        sp = k
        if p.annotation != Parameter.empty:
            sp = f"{sp}:{str(p.annotation)}"
        if p.default != Parameter.empty:
            sp = f"{sp}={p.default}"
        sig = f"{sp}" if i == 1 else f"{sig}, {sp}"
    return f"({sig})"


def positional_params(func):
    res = ''
    params = inspect.signature(func).parameters
    for i, (k, p) in enumerate(params.items()):
        if i == 0 or p.kind not in [Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD,
                                    Parameter.VAR_POSITIONAL]:
            continue
        if p.kind == Parameter.POSITIONAL_OR_KEYWORD and p.default != Parameter.empty:
            continue
        sp = f"'{k}' : {k}"
        res = f"{sp}" if i == 1 else f"{res}, {sp}"
    return res


def keyword_params(func):
    res = ''
    params = inspect.signature(func).parameters
    for i, (k, p) in enumerate(params.items()):
        if i == 0 or p.kind not in [Parameter.KEYWORD_ONLY, Parameter.POSITIONAL_OR_KEYWORD, Parameter.VAR_KEYWORD]:
            continue
        if p.kind == Parameter.POSITIONAL_OR_KEYWORD and p.default == Parameter.empty:
            continue
        sp = f"'{k}' : {k}"
        res = f"{sp}" if i == 1 else f"{res}, {sp}"
    return res



import base64
import functools
import hashlib
import importlib
import sysconfig
import tempfile
from pathlib import Path
from shutil import copyfile, copytree, ignore_patterns, make_archive

import click

from .command import CwsCommand
from .error import CwsCommandError
from .. import aws


class CwsZipArchiver(CwsCommand):
    """
    This command uploads project source folder as a zip file on a S3 bucket.
    Uploads also the hash code of this file to be able to determined code changes (used by terraform as a trigger).
    """

    @property
    def options(self):
        return [
            *super().options,
            click.option('--bucket', '-b', help="Bucket to upload sources zip file to", required=True),
            click.option('--debug', is_flag=True, help="Print debug logs to stderr."),
            click.option('--dry', is_flag=True, help="Doesn't perform upload."),
            click.option('--hash', is_flag=True, help="Upload also hash code content."),
            click.option('--ignore', '-i', multiple=True, help="Ignore pattern."),
            click.option('--key', '-k', help="Sources zip file bucket's name."),
            click.option('--module_name', '-m', multiple=True, help="Python module added from current pyenv."),
            click.option('--profile_name', '-p', required=True, help="AWS credential profile."),
        ]

    @classmethod
    def multi_execute(cls, project_dir, workspace, execution_list):
        for command, options in execution_list:
            command._zip(**options)

    def __init__(self, app=None, name='zip'):
        super().__init__(app, name=name)

    def _zip(self, *, project_dir, module, bucket, key, profile_name, module_name, hash, dry, debug, ignore, **options):
        aws_s3_session = aws.AwsS3Session(profile_name=profile_name)
        module_name = module_name or []

        key = key if key else f"{module}-{self.app.name}"
        if debug:
            name = f"{module}-{options['service']}"
            where = f"{bucket}/{key}"
            print(f"Uploading zip sources of {name} at s3:{where} {'(not done)' if dry else ''}")

        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_path = Path(tmp_dir)

            if ignore and type(ignore) is not list:
                ignore = [ignore]
            full_ignore_patterns = functools.partial(ignore_patterns, '*.pyc', '__pycache__', 'bin', 'test', *ignore)

            # Creates archive
            copytree(project_dir, str(tmp_path / 'filtered_dir'),
                     ignore=full_ignore_patterns('*cws.yml', 'env_variables*'))
            for name in module_name:
                if name.endswith(".py"):
                    file_path = Path(sysconfig.get_path('purelib')) / name
                    copyfile(file_path, str(tmp_path / f'filtered_dir/{name}'))
                else:
                    mod = importlib.import_module(name)
                    module_path = Path(mod.__file__).resolve().parent
                    copytree(module_path, str(tmp_path / f'filtered_dir/{name}'), ignore=full_ignore_patterns())
            module_archive = make_archive(str(tmp_path / 'sources'), 'zip', str(tmp_path / 'filtered_dir'))

            # Uploads archive on S3
            with open(module_archive, 'rb') as module_archive:
                b64sha256 = base64.b64encode(hashlib.sha256(module_archive.read()).digest())
                module_archive.seek(0)
                try:
                    if not dry:
                        if debug:
                            print(f"Upload sources...")
                        aws_s3_session.client.upload_fileobj(module_archive, bucket, key)
                        if debug:
                            print(f"Successfully uploaded sources at s3://{bucket}/{key}")
                except Exception as e:
                    print(f"Failed to upload module sources on S3 : {e}")
                    raise CwsCommandError(str(e))

            # Creates hash value
            if hash:
                with tmp_path.with_name('b64sha256_file').open('wb') as b64sha256_file:
                    b64sha256_file.write(b64sha256)

                # Uploads archive hash value to bucket
                with tmp_path.with_name('b64sha256_file').open('rb') as b64sha256_file:
                    try:
                        if not dry:
                            if debug:
                                print(f"Upoad sources hash...")
                            aws_s3_session.client.upload_fileobj(b64sha256_file, bucket, f"{key}.b64sha256",
                                                                 ExtraArgs={'ContentType': 'text/plain'})
                            if debug:
                                print(f"Successfully uploaded sources hash at s3://{bucket}/{key}.b64sha256")
                    except Exception as e:
                        print(f"Failed to upload archive hash on S3 : {e}")
                        raise CwsCommandError(str(e))



from dataclasses import dataclass

import json
import os
import re
import requests
from chalice import AuthResponse
from chalice import Chalice, Blueprint as ChaliceBlueprint
from chalice.app import AuthRequest
from typing import Callable
from typing import Dict, List, Union, Optional

from .config import Config, LocalConfig, DevConfig, ProdConfig, DEFAULT_PROJECT_DIR, DEFAULT_WORKSPACE
from .mixins import CoworksMixin
from .utils import trim_underscores, HTTP_METHODS

ENTRY_REGEXP = '[^0-9a-zA-Z_]'


#
# Decorators
#

def entry(fun):
    """Decorator to create a microservice entrypoint from function name."""
    name = fun.__name__.upper()
    for method in HTTP_METHODS:
        if name == method:
            fun.__CWS_METHOD = method
            fun.__CWS_PATH = ''
            return fun
        if name.startswith(f'{method}_'):
            fun.__CWS_METHOD = method
            name = fun.__name__[len(method) + 1:]
            name = trim_underscores(name)  # to allow several functions with different args
            fun.__CWS_PATH = name.replace('_', '/')
            return fun
    raise AttributeError(f"The function name {fun.__name__} doesn't start with a HTTP method name.")


def hide(f):
    """Hide a route of the microservice.

     May be used as a decorator.

     Usefull when creating inherited microservice.
     """

    setattr(f, '__cws_hidden', True)
    return f


#
# Classes
#

@dataclass
class Entry:
    """An entry is an API entry defined on a microservice, with a specific authorization function and
    its response function."""

    auth: Callable
    fun: Callable


@dataclass
class ScheduleEntry:
    """An schedule entry is an EventBridge entry defined on a microservice, with the schedule expression,
    its description and its response function."""

    name: str
    exp: str
    desc: str
    fun: Callable


class Blueprint(CoworksMixin, ChaliceBlueprint):
    """ Represents a blueprint, list of routes that will be added to microservice when registered.

    See :ref:`Blueprint <blueprint>` for more information.

    """

    def __init__(self, name=None, **kwargs):
        """Initialize a blueprint.

        :param kwargs: Other Chalice parameters.

        """
        import_name = name or self.__class__.__name__.lower()
        super().__init__(import_name)

    @property
    def name(self):
        return self._import_name

    @property
    def current_app(self):
        return self._current_app

    @property
    def logger(self):
        return self._current_app.logger

    def entry(self, route):
        return self.current_app.entry(route)

    def deferred_init(self, workspace):
        for func in self.before_first_activation_funcs:
            self.current_app.before_first_activation(func)
        for func in self.before_activation_funcs:
            self.current_app.before_activation(func)
        for func in self.after_activation_funcs:
            self.current_app.after_activation(func)
        for func in self.handle_exception_funcs:
            self.current_app.handle_exception(func)


class TechMicroService(CoworksMixin, Chalice):
    """Simple tech microservice.
    
    See :ref:`tech` for more information.
    
    """

    def __init__(self, name: str = None, configs: Union[Config, List[Config]] = None, **kwargs):
        """ Initialize a technical microservice.
        :param name: Name used to identify the microservice.
        :param configs: Deployment configurations.
        :param kwargs: Other Chalice parameters.
        """
        name = name or self.__class__.__name__.lower()

        self.configs = configs or [LocalConfig(), DevConfig(), ProdConfig()]
        if type(self.configs) is not list:
            self.configs = [configs]
        self.config = None

        super().__init__(app_name=name, **kwargs)

        # Blueprints and extended commands added
        self.experimental_feature_flags.update([
            'BLUEPRINTS'
        ])
        self.blueprints: Dict[str, Blueprint] = {}
        self.commands = {}
        self.entries: Optional[Dict[str, Dict[str, Entry]]] = None

        # App init deferered functions.
        self.deferred_inits = []
        self._got_first_activation = False

    @property
    def name(self):
        return self.app_name

    @property
    def ms_type(self):
        return 'tech'

    def get_config(self, workspace):
        for conf in self.configs:
            if conf.is_valid_for(workspace):
                return conf
        return Config()

    def deferred_init(self, workspace):
        if self.entries is None:

            # Set workspace config
            self.config = self.get_config(workspace)

            # Initializes routes and defered initializations
            self._init_routes(self)
            for deferred_init in self.deferred_inits:
                deferred_init(workspace)
            for blueprint in self.blueprints.values():
                blueprint.deferred_init(workspace)

    def register_blueprint(self, blueprint: Blueprint, name=None, url_prefix='', hide_routes=False, show_routes=None):
        """ Register a :class:`Blueprint` on the microservice.

        :param blueprint: blueprint to register.
        :param name: name registration (needed to retrieve it).
        :param url_prefix: url prefix for the routes added (must be unique for each blueprints).
        :param hide_routes: Hide all routes in blueprint.
        :param show_routes:
        :return:
        """
        name = name or blueprint.name
        if name in self.blueprints:
            raise NameError(f"A blueprint is already defined with the name : {name}.")
        self.blueprints[name] = blueprint

        def deferred(workspace):
            if not hide_routes:
                blueprint._init_routes(self, url_prefix=url_prefix)
            super(TechMicroService, self).register_blueprint(blueprint, url_prefix=url_prefix)

        self.deferred_inits.append(deferred)

    def add_entry(self, path, method, auth, fun):
        self.entries[path][method] = Entry(auth, fun)

    def execute(self, command, *, project_dir=DEFAULT_PROJECT_DIR, module=None, service=None,
                workspace=DEFAULT_WORKSPACE, output=None, error=None, **options):
        from .cws.client import CwsClientOptions
        from .cws.error import CwsCommandError

        # Executes a coworks command.
        module = __name__ if module is None else module
        service = self.name if service is None else service
        cws_options = CwsClientOptions({"project_dir": project_dir, 'module': module, 'service': service})
        service_config = cws_options.get_service_config(module, service, workspace)
        if type(command) is str:
            cmd = service_config.get_command(command, self)
            if not cmd:
                raise CwsCommandError(f"The command {command} was not added to the microservice {self.name}.\n")
        else:
            cmd = command
        command_options = service_config.get_command_options(command)
        execute_options = {**command_options, **options}
        cmd.execute(output=output, error=error, **execute_options)

    def __call__(self, event, context):
        """Lambda handler."""
        try:
            self.do_before_first_activation(event, context)
            self.do_before_activation(event, context)
            response = self.handler(event, context)
            self.do_after_activation(response)
            return self._convert_response(response)
        except Exception as e:
            self.log.error(f"exception: {e}")
            response = self.do_handle_exception(event, context, e)
            if response is None:
                raise
            return self._convert_response(response)

    def handler(self, event, context):
        """Main microservice entry point."""

        if event.get('type') == 'TOKEN':
            return self._token_handler(event, context)
        return self._api_handler(event, context)

    def deferred(self, f):
        """Registers a function to be run once the microservice will be initialized.

        May be used as a decorator.

        The function will be called with one parameter, the workspace, and its return value is ignored.
        """

        self.deferred_inits.append(f)
        return f

    def do_before_first_activation(self, event, context):
        """Calls all before first activation functions."""
        if self._got_first_activation:
            return

        workspace = os.environ['WORKSPACE']
        self.deferred_init(workspace=workspace)

        if not self._got_first_activation:
            for func in self.before_first_activation_funcs:
                func(event, context)
            self._got_first_activation = True

    def do_before_activation(self, event, context):
        """Calls all before activation functions."""
        for func in self.before_activation_funcs:
            func(event, context)

    def do_after_activation(self, response):
        """Calls all after activation functions."""
        for func in reversed(self.after_activation_funcs):
            resp = func(response)
            if resp is not None:
                response = resp
        return response

    def do_handle_exception(self, event, context, exc):
        """Calls all exception handlers."""
        for func in reversed(self.handle_exception_funcs):
            resp = func(event, context, exc)
            if resp is not None:
                return resp

    def _entry(self, path: str, method: str) -> Optional[Entry]:
        """Finds the entry corresponding to the route."""
        pathes = [x for x in path.split('/') if x]
        for entry_path, result in self.entries.items():
            entry_pathes = [x for x in entry_path.split('/') if x]
            if len(pathes) != len(entry_pathes):
                continue

            found = True
            for index, path in enumerate(entry_pathes):
                if index < len(pathes):
                    if path.startswith('{') or path == pathes[index]:
                        continue
                found = False
                break

            if found:
                return result[method]
        return None

    def _token_handler(self, event, context):
        """Authorization handler."""
        self.log.debug(f"Calling {self.name} for authorization")

        try:
            *_, method, route = event.get('methodArn').split('/', 3)
            authorizer = self._entry(f'/{route}', method).auth
            return authorizer(event, context)
        except Exception as e:
            self.log.debug(f"Error in authorization handler for {self.name} : {e}")
            request = AuthRequest(event['type'], event['authorizationToken'], event['methodArn'])
            return AuthResponse(routes=[], principal_id='user').to_dict(request)

    def _api_handler(self, event, context):
        """API rest handler."""
        self.log.debug(f"Calling {self.name} by api")

        try:
            # Chalice accepts only string for body
            if type(event['body']) is dict:
                event['body'] = json.dumps(event['body'])

            res = super().__call__(event, context)
            workspace = os.getenv('WORKSPACE')
            if workspace:
                res['headers']['x-cws-workspace'] = workspace

            self.log.debug(f"Call {self.name} returns {res}")
            return res
        except Exception as e:
            self.log.debug(f"Error in api handler for {self.name} : {e}")
            raise


class MicroServiceProxy:

    def __init__(self, env_name, **kwargs):
        self.session = requests.Session()
        self.cws_id = os.getenv(f'{env_name}_CWS_ID')
        self.cws_token = os.getenv(f'{env_name}_CWS_TOKEN')
        self.cws_stage = os.getenv(f'{env_name}_CWS_STAGE')

        self.session.headers.update(
            {
                'authorization': self.cws_token,
                'content-type': 'application/json',
            })
        self.url = f"https://{self.cws_id}.execute-api.eu-west-1.amazonaws.com/{self.cws_stage}"

    def get(self, path, data=None):
        return self.session.get(f'{self.url}/{path}', data=data)

    def post(self, path, data=None, attachments=None, headers=None, sync=True):
        if headers:
            self.session.headers.update(headers)
        if not sync:
            self.session.headers.update({'InvocationType': 'Event'})
        return self.session.post(f'{self.url}/{path}', json=data or {}, files=attachments)


class BizMicroService(TechMicroService):
    """Biz composed microservice activated by events.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.schedule_entries: Dict[str, ScheduleEntry] = {}

    @property
    def ms_type(self):
        return 'biz'

    # noinspection PyMethodOverriding
    def schedule(self, exp, *, name=None, description=None, workspaces=None):
        """Registers a a schedule event to trigger the microservice.

        May be used as a decorator.

        The function will be called with the event name.
        """

        def decorator(f):
            schedule_name = name if name else exp
            desc = description if description else f.__doc__
            entry_key = re.sub(ENTRY_REGEXP, '', exp)
            self.schedule_entries[f"{f.__name__}_{entry_key}"] = ScheduleEntry(schedule_name, exp, desc, f)
            return f

        return decorator

    def handler(self, event, context):
        """Main microservice entry point."""

        if event.get('type') == 'CWS_SCHEDULE_EVENT':
            return self._schedule_event_handler(event, context)
        return super().handler(event, context)

    def _schedule_event_handler(self, event, context):
        """Schedule event handler."""
        self.log.debug(f"Calling {self.name} by event bridge")

        try:
            entry_name = event.get('entry_name')
            return self.schedule_entries[entry_name].fun(event.get('schedule_name'))
        except Exception as e:
            self.log.debug(f"Error in schedule event handler for {self.name} : {e}")
            raise



import json
import os
from typing import List, Tuple, Union, Any

import requests
from aws_xray_sdk.core import xray_recorder

from .. import Blueprint, entry
from ..error import NotFoundError, InternalServerError

Response = Tuple[dict, int]
GetResponse = Tuple[Union[dict, List[dict]], int]
Ids = List[int]
Filters = List[Tuple[str, str, Any]]


class AccessDenied(Exception):
    ...


class Odoo(Blueprint):
    """Odoo blueprint.
    Environment variables needed:
    - env_url_var_name: Variable name for the odoo server URL.
    - env_dbname_var_name: Variable name for the odoo database.
    - env_user_var_name: Variable name for the user login.
    - env_passwd_var_name: Variable name for the password.
    """

    def __init__(self,
                 env_url_var_name=None, env_dbname_var_name=None, env_user_var_name=None, env_passwd_var_name=None,
                 env_var_prefix=None, **kwargs):
        super().__init__(**kwargs)
        if env_var_prefix:
            self.env_url_var_name = f"{env_var_prefix}_URL"
            self.env_dbname_var_name = f"{env_var_prefix}_DBNAME"
            self.env_user_var_name = f"{env_var_prefix}_USER"
            self.env_passwd_var_name = f"{env_var_prefix}_PASSWD"
        else:
            self.env_url_var_name = env_url_var_name
            self.env_dbname_var_name = env_dbname_var_name
            self.env_user_var_name = env_user_var_name
            self.env_passwd_var_name = env_passwd_var_name
        self.url = self.dbname = self.user = self.passwd = None
        self.session_id = None

        @self.before_first_activation
        def check_env_vars(event, context):
            self.url = os.getenv(self.env_url_var_name)
            if not self.url:
                raise EnvironmentError(f'{self.env_url_var_name} not defined in environment.')
            self.dbname = os.getenv(self.env_dbname_var_name)
            if not self.dbname:
                raise EnvironmentError(f'{self.env_dbname_var_name} not defined in environment.')
            self.user = os.getenv(self.env_user_var_name)
            if not self.user:
                raise EnvironmentError(f'{self.env_user_var_name} not defined in environment.')
            self.passwd = os.getenv(self.env_passwd_var_name)
            if not self.passwd:
                raise EnvironmentError(f'{self.env_passwd_var_name} not defined in environment.')

        @self.before_activation
        def set_session(event, context):
            self.session_id, status_code = self._get_session_id()

        @self.after_activation
        def destroy_session(response):
            self.session_id = self._destroy_session()

        @self.handle_exception
        def handle_access_denied(event, context, e):
            if type(e) is AccessDenied:
                return "AccessDenied", 401
            raise

    @entry
    def get(self, model: str, query: str = "{*}", order: str = None, filters: Filters = None,
            limit: int = 300, page_size=None, page=0, ensure_one=False) -> GetResponse:
        params = {'query': query, 'limit': limit}
        if order:
            params.update({'order': order})
        if filters:
            params.update({'filters': json.dumps(filters)})
        if page_size:
            params.update({'page_size': page_size})
        if page:
            params.update({'page': page})
        res, status_code = self.odoo_get(f'{self.url}/api/{model}', params)
        if status_code == 200:
            if ensure_one:
                if len(res['result']) == 0:
                    raise NotFoundError(f"Nothing found for {model} with {filters} [Odoo blueprint {self.name}]")
                if len(res['result']) > 1:
                    raise InternalServerError("More than one result")
                else:
                    return res['result'][0], 200
            return res['result'], 200
        if status_code == 500:
            raise InternalServerError(f"{res}  [Odoo blueprint {self.name}]")
        return res, status_code

    @entry
    def get_(self, model: str, rec_id: int, query="{*}") -> Response:
        params = {'query': query}
        res, status_code = self.odoo_get(f'{self.url}/api/{model}/{rec_id}', params)
        if status_code == 500:
            raise InternalServerError(f"{res}  [Odoo blueprint {self.name}]")
        return res, status_code

    @entry
    def get_pdf(self, report_id: int, rec_ids: Ids) -> Response:
        params = {'params': {'res_ids': json.dumps(rec_ids)}}
        res, status_code = self.odoo_post(f"{self.url}/report/{report_id}", params=params)
        if status_code == 200:
            return res['result'], 200
        if status_code == 500:
            raise InternalServerError(f"{res}  [Odoo blueprint {self.name}]")
        return res, status_code

    @entry
    def post(self, model: str, data=None, context=None) -> Response:
        params = {'params': {'data': data or {}}}
        if context:
            params.update({'context': context})
        res, status_code = self.odoo_post(f"{self.url}/api/{model}", params=params)
        if status_code == 200:
            return res['result'], 200
        if status_code == 500:
            raise InternalServerError(f"{res}  [Odoo blueprint {self.name}]")
        return res, status_code

    @entry
    def put(self, model: str, rec_id: int, data=None) -> Response:
        params = {'params': {'data': data or {}}}
        res, status_code = self.odoo_put(f'{self.url}/api/{model}/{rec_id}', params)
        if status_code == 500:
            raise InternalServerError(f"{res}  [Odoo blueprint {self.name}]")
        return res, status_code

    @entry
    def put_(self, model: str, filters: Filters = None, data=None) -> Response:
        """Bulk update."""
        params = {'params': {'data': data or {}}}
        if filters:
            params.update({'filter': filters})
        res, status_code = self.odoo_put(f'{self.url}/api/{model}', params)
        if status_code == 500:
            raise InternalServerError(f"{res}  [Odoo blueprint {self.name}]")
        return res, status_code

    @entry
    def delete(self, model: str, rec_id: int) -> Response:
        res, status_code = self.odoo_delete(f'{self.url}/api/{model}/{rec_id}')
        if status_code == 500:
            raise InternalServerError(f"{res}  [Odoo blueprint {self.name}]")
        return res, status_code

    @xray_recorder.capture()
    def odoo_get(self, path, params, headers=None):
        params.update({'jsonrpc': "2.0", 'session_id': self.session_id})
        headers = headers or {}
        res = requests.get(path, params=params, headers=headers)
        try:
            result = res.json()
            if 'error' in result:
                return f"{result['message']}: {result['data']}", 404
            return result, res.status_code
        except (json.decoder.JSONDecodeError, Exception):
            return res.text, 500

    @xray_recorder.capture()
    def odoo_post(self, path, params, headers=None) -> Tuple[Union[str, dict], int]:
        _params = {'jsonrpc': "2.0", 'session_id': self.session_id}
        headers = headers or {}
        res = requests.post(path, params=_params, json=params, headers=headers)
        try:
            result = res.json()
            if 'error' in result:
                return f"{result['error']['message']}:{result['error']['data']}", 404
            return result, res.status_code
        except (json.decoder.JSONDecodeError, Exception):
            return res.text, 500

    @xray_recorder.capture()
    def odoo_put(self, path, params, headers=None) -> Tuple[Union[str, dict], int]:
        _params = {'jsonrpc': "2.0", 'session_id': self.session_id}
        headers = headers or {}
        res = requests.put(path, params=_params, json=params, headers=headers)
        try:
            result = res.json()
            if 'error' in result:
                return f"{result['error']['message']}:{result['error']['data']}", 404
            return result, res.status_code
        except (json.decoder.JSONDecodeError, Exception):
            return res.text, 500

    @xray_recorder.capture()
    def odoo_delete(self, path, headers=None) -> Tuple[Union[str, dict], int]:
        _params = {'jsonrpc': "2.0", 'session_id': self.session_id}
        headers = headers or {}
        res = requests.delete(path, params=_params, headers=headers)
        try:
            result = res.json()
            if 'error' in result:
                return f"{result['error']['message']}:{result['error']['data']}", 404
            return result, res.status_code
        except (json.decoder.JSONDecodeError, Exception):
            return res.text, 500

    def _get_session_id(self):
        """Open or checks the connection."""
        try:
            data = {
                'jsonrpc': "2.0",
                'params': {
                    'db': self.dbname,
                    'login': self.user,
                    'password': self.passwd,
                }
            }
            res = requests.post(
                f'{self.url}/web/session/authenticate/',
                data=json.dumps(data),
                headers={'Content-type': 'application/json'}
            )

            data = json.loads(res.text)
            if data['result']['session_id']:
                return res.cookies["session_id"], 200
        except Exception as e:
            raise AccessDenied()

    def _destroy_session(self):
        """Open or checks the connection."""
        try:
            data = {
                'jsonrpc': "2.0",
                'params': {
                    'db': self.dbname,
                    'login': self.user,
                    'password': self.passwd,
                }
            }
            res = requests.post(
                f'{self.url}/web/session/destroy/',
                data=json.dumps(data),
                headers={'Content-type': 'application/json'}
            )

            return "", 200
        except Exception as e:
            return str(e), 401



import os

from coworks import Blueprint
import redis


class RedisBlueprint(Blueprint):

    def __init__(self, endpoint_env_var_name, passwd_env_var_name, port_env_var_name, **kwargs):
        super().__init__(**kwargs)
        self.redis = None

        @self.before_first_activation
        def load_env_var(*args, **kwargs):
            endpoint = os.getenv(endpoint_env_var_name)
            passwd = os.getenv(passwd_env_var_name)
            port = os.getenv(port_env_var_name)
            self.redis = redis.Redis(host=endpoint, password=passwd, port=port, db=0)



import itertools
import logging
import sys
from abc import ABC
from dataclasses import dataclass
from itertools import chain, repeat
from pathlib import Path
from threading import Thread
from time import sleep
from typing import List, Optional

import boto3
import click

from coworks.aws import AwsS3Session
from coworks.coworks import Entry
from .command import CwsCommand, CwsCommandError
from .writer import CwsTemplateWriter
from .zip import CwsZipArchiver
from ..config import CORSConfig

UID_SEP = '_'


@dataclass
class TerraformResource:
    parent_uid: str
    path: str
    entries: List[Entry]
    cors: CORSConfig

    @property
    def uid(self):
        def remove_brackets(path):
            return f"{path.replace('{', '').replace('}', '')}"

        if self.path is None:
            return ''

        last = remove_brackets(self.path)
        return f"{self.parent_uid}{UID_SEP}{last}" if self.parent_uid else last

    @property
    def is_root(self):
        return self.path is None

    @property
    def parent_is_root(self):
        return self.parent_uid == ''

    def __repr__(self):
        return f"{self.uid}:{self.entries}"


class CwsTerraformCommand(CwsCommand, ABC):
    WRITER_CMD = 'export'

    @property
    def options(self):
        return [
            *super().options,
            click.option('--memory_size', default=128),
            click.option('--timeout', default=30),
        ]

    def __init__(self, app=None, **kwargs):
        self.writer_cmd = self.add_writer_command(app)
        super().__init__(app, **kwargs)

    def add_writer_command(self, app):
        """Default writer command added if not already defined."""
        return app.commands.get(self.WRITER_CMD) or CwsTemplateWriter(app)

    @classmethod
    def generate_terraform_files(cls, step, app, terraform, filename, msg, **options):
        debug = options['debug']
        dry = options['dry']
        profile_name = options['profile_name']
        aws_region = boto3.Session(profile_name=profile_name).region_name

        if not dry or options.get('stop') == step:
            if debug:
                print(msg)
            output = str(Path(terraform.working_dir) / filename)
            app.execute(cls.WRITER_CMD, template=["terraform.j2"], output=output, aws_region=aws_region,
                        step=step, resources=cls.terraform_resources(app), **options)

    @staticmethod
    def terraform_resources(app):
        """Returns the list of flatten path (prev, last, entry)."""
        resources = {}

        def add_entries(previous, last, entries: Optional[List[Entry]]):
            ter_entry = TerraformResource(previous, last, entries, app.config.cors)
            uid = ter_entry.uid
            if uid not in resources:
                resources[uid] = ter_entry
            if resources[uid].entries is None:
                resources[uid].entries = entries
            return uid

        for route, entries in app.entries.items():
            previous_uid = ''
            if route.startswith('/'):
                route = route[1:]
            splited_route = route.split('/')

            # special root case
            if splited_route == ['']:
                add_entries(None, None, entries)
                continue

            # creates intermediate resources
            last_path = splited_route[-1:][0]
            for prev in splited_route[:-1]:
                previous_uid = add_entries(previous_uid, prev, None)

            # set entry keys for last entry
            add_entries(previous_uid, last_path, entries)

        return resources


class CwsTerraformDeployer(CwsTerraformCommand):
    """ Deploiement in 4 steps:
    create
        Step 1. Create API in default workspace (destroys API integrations made in previous deployment)
        Step 2. Create Lambda in stage workspace (destroys API deployment made in previous deployment)
    update
        Step 3. Update API routes integrations
        Step 4. Update API deployment
    """

    ZIP_CMD = 'zip'

    @property
    def options(self):
        return [
            *super().options,
            *self.zip_cmd.options,
            click.option('--binary_media_types'),
            click.option('--create', '-c', is_flag=True, help="May create or recreate the API [Global option only]."),
            click.option('--dry', is_flag=True, help="Doesn't perform deploy [Global option only]."),
            click.option('--layers', '-l', multiple=True),
            click.option('--output', '-o', is_flag=True, help="Print terraform output values."),
            click.option('--stop', type=click.Choice(['create', 'update']), help="Stop the terraform generation"),
            click.option('--update', '-u', is_flag=True, help="Only update lambda code [Global option only]."),
        ]

    @classmethod
    def multi_execute(cls, project_dir, workspace, execution_list):
        terraform = Terraform()
        terraform.init()

        # Output, dry, create and update are global options
        dry = create = update_lambda_only = False
        for command, options in execution_list:
            dry = options.pop('dry', False) or dry
            create = options.pop('create', False) or create
            update_lambda_only = options.pop('update', False) or update_lambda_only

            # Set default bucket key value
            options['key'] = options['key'] or f"{options.get('module')}-{command.app.name}/archive.zip"

            # Stop if only print output
            output = options.pop('output', False)
            if output:
                print(f"terraform output : {terraform.output()}")
                return

        # Validates create option choice
        if create:
            prompts = chain(["Are you sure you want to (re)create the API [yN]?:"], repeat("Answer [yN]: "))
            replies = map(input, prompts)
            valid_response = next(filter(lambda x: x == 'y' or x == 'n' or x == '', replies))
            if valid_response != 'y':
                return

        # Transfert zip file to S3 (to be done on each service)
        for command, options in execution_list:
            print(f"Uploading zip to S3")
            module_name = options.pop('module_name')
            ignore = options.pop('ignore') or ['.*', 'terraform']
            options.pop('hash')
            command.app.execute(cls.ZIP_CMD, ignore=ignore, module_name=module_name, hash=True, **options)

        # Generates default provider
        cls.generate_common_terraform_files()

        # Generates terraform files (create step)
        if not update_lambda_only:
            for command, options in execution_list:
                terraform_filename = f"{command.app.name}.{command.app.ms_type}.tf"
                msg = f"Generate terraform files for creating API and lambdas for {command.app.name}"
                cls.generate_terraform_files("create", command.app, terraform, terraform_filename, msg, dry=dry,
                                             **options)

        # Apply terraform if not dry (create API with null resources and lambda step)
        # or in case of only updating lambda code
        if not dry:
            msg = ["Create API", "Create lambda"] if create else ["Update API", "Update lambda"]
            cls.terraform_apply(terraform, workspace, msg, update_lambda_only=update_lambda_only)

        # Generates terraform files (update step)
        if not update_lambda_only:
            for command, options in execution_list:
                terraform_filename = f"{command.app.name}.{command.app.ms_type}.tf"
                msg = f"Generate terraform files for updating API routes and deploiement for {command.app.name}"
                cls.generate_terraform_files("update", command.app, terraform, terraform_filename, msg, dry=dry,
                                             **options)

            # Apply terraform if not dry (update API routes and deploy step)
            if not dry:
                msg = ["Update API routes", f"Deploy API {workspace}"]
                cls.terraform_apply(terraform, workspace, msg)

        # Traces output
        print(f"terraform output : {terraform.output()}")

    def __init__(self, app=None, name='deploy'):
        self.zip_cmd = self.add_zip_command(app)
        super().__init__(app, name=name)

    def add_zip_command(self, app):
        """Default zip command added if not already defined."""
        return app.commands.get(self.ZIP_CMD) or CwsZipArchiver(app)

    @staticmethod
    def generate_common_terraform_files():
        with open('terraform/default_provider.tf', 'w') as output:
            print('provider "aws" {\nprofile = "fpr-customer"\nregion = "eu-west-1"\n}', file=output, flush=True)

    @staticmethod
    def terraform_apply(terraform, workspace, traces, update_lambda_only=False):
        """In the default terraform workspace, we have the API.
        In the specific workspace, we have the corresponding stagging lambda.
        """
        stop = False

        def display_spinning_cursor():
            spinner = itertools.cycle('|/-\\')
            while not stop:
                sys.stdout.write(next(spinner))
                sys.stdout.write('\b')
                sys.stdout.flush()
                sleep(0.1)

        spin_thread = Thread(target=display_spinning_cursor)
        spin_thread.start()

        try:
            if not update_lambda_only:
                print(f"Terraform apply ({traces[0]})", flush=True)
                terraform.apply("default")
            print(f"Terraform apply ({traces[1]})", flush=True)
            terraform.apply(workspace)
        finally:
            stop = True


class CwsTerraformDestroyer(CwsTerraformCommand):

    @property
    def options(self):
        return [
            *super().options,
            click.option('--bucket', '-b', help="Bucket to remove sources zip file from", required=True),
            click.option('--debug', is_flag=True, help="Print debug logs to stderr."),
            click.option('--dry', is_flag=True, help="Doesn't perform destroy."),
            click.option('--key', '-k', help="Sources zip file bucket's name."),
            click.option('--profile_name', '-p', required=True, help="AWS credential profile."),
        ]

    @classmethod
    def multi_execute(cls, project_dir, workspace, execution_list):
        for command, options in execution_list:
            command.rm_zip(**options)
            command.terraform_destroy(**options)

    def __init__(self, app=None, name='destroy'):
        super().__init__(app, name=name)

    def rm_zip(self, *, module, bucket, key, profile_name, dry, debug, **options):
        aws_s3_session = AwsS3Session(profile_name=profile_name)

        # Removes zip file from S3
        key = key if key else f"{module}-{self.app.name}"
        if debug:
            name = f"{module}-{options['service']}"
            where = f"{bucket}/{key}"
            print(f"Removing zip sources of {name} from s3: {where} {'(not done)' if dry else ''}")

        if not dry:
            aws_s3_session.client.delete_object(Bucket=bucket, Key=key)
            aws_s3_session.client.delete_object(Bucket=bucket, Key=f"{key}.b64sha256")
            if debug:
                print(f"Successfully removed sources at s3://{bucket}/{key}")

    def terraform_destroy(self, *, workspace, debug, dry, **options):
        terraform = Terraform()

        if not dry:
            for w in terraform.workspace_list():
                if w in ["default", workspace]:
                    print(f"Terraform destroy ({w})", flush=True)
                    terraform.destroy(w)

        terraform_filename = f"{self.app.name}.{self.app.ms_type}.tf"
        output = Path(terraform.working_dir) / terraform_filename
        if debug:
            print(f"Removing terraform files: {output} {'(not done)' if dry else ''}")
        if not dry:
            output.unlink(missing_ok=True)


logging.getLogger("python_terraform").setLevel(logging.ERROR)


class Terraform:

    def __init__(self):
        from python_terraform import Terraform as PythonTerraform

        self.terraform = PythonTerraform(working_dir='terraform', terraform_bin_path='terraform')
        dir = Path(self.working_dir).mkdir(exist_ok=True)
        return_code, _, err = self.terraform.init(dir_or_plan=dir)
        if return_code != 0:
            raise CwsCommandError(err)

    @property
    def working_dir(self):
        return self.terraform.working_dir

    def init(self):
        return_code, _, err = self.terraform.init()
        if return_code != 0:
            raise CwsCommandError(err)

    def apply(self, workspace):
        self.select_workspace(workspace)
        return_code, _, err = self.terraform.apply(skip_plan=True, input=False, raise_on_error=False, parallelism=1)
        if return_code != 0:
            raise CwsCommandError(err)

    def destroy(self, workspace):
        self.select_workspace(workspace)
        return_code, _, err = self.terraform.destroy()
        if return_code != 0:
            raise CwsCommandError(err)

    def output(self):
        self.select_workspace("default")
        values = self.terraform.output(capture_output=True)
        return {key: value['value'] for key, value in values.items()} if values else "{}"

    def workspace_list(self):
        self.select_workspace("default")
        return_code, out, err = self.terraform.cmd('workspace', 'list')
        if return_code != 0:
            raise CwsCommandError(err)
        values = out[1:].translate(str.maketrans('', '', ' \t\r')).split('\n')
        return filter(None, values)

    def select_workspace(self, workspace):
        return_code, out, err = self.terraform.workspace('select', workspace)
        if workspace != 'default' and return_code != 0:
            _, out, err = self.terraform.workspace('new', workspace, raise_on_error=True)
        if not (Path(self.working_dir) / '.terraform').exists():
            self.terraform.init(input=False, raise_on_error=True)



import json
import urllib.parse
from json import JSONDecodeError

import jinja2
from chalice import NotFoundError

from coworks import TechMicroService, entry


class JinjaRenderMicroService(TechMicroService):
    """ Render a jinja template to html
        Templates can be sent to the microservice in 2 differents ways :
            - send files in multipart/form-data body
            - put template content in url """

    def __init__(self, globals=None, filters=None, **kwargs):
        super().__init__(**kwargs)
        self.globals = globals
        self.filters = filters

    def create_environment(self, loader):
        env = jinja2.Environment(loader=loader)
        if self.globals:
            env.globals.update(**self.globals)
        if self.filters:
            env.filters.update(**self.filters)
        return env

    @entry
    def post_render(self, template_to_render_name, templates=None, context=None):
        """ render the template named template_to_render_name using templates sources given in templates
            pass templates as files of a multipart/form-data body
            pass context as a json file """
        if self.debug:
            print("template_to_render_name", template_to_render_name)
            print("templates :", templates)
            print("context :", context)

        if templates is None:
            raise NotFoundError("At least one template is expected")

        templates = [templates] if not isinstance(templates, list) else templates
        context = context if context is not None else {}
        templates_dict = {template.file.name: template.file.read().decode('utf-8-sig') for template in templates}

        if self.debug:
            print("templates_dict", templates_dict)

        env = self.create_environment(loader=jinja2.DictLoader(templates_dict))
        template_to_render = env.get_template(template_to_render_name)
        render = template_to_render.render(**context)

        try:
            json_render = json.loads(render)
            return json_render
        except JSONDecodeError as e:
            if self.debug:
                print(e)
            return {"render": render}

    @entry
    def get_render_(self, template):
        """ render template which content is given in url
        pass jinja context in query_params """
        template = urllib.parse.unquote_plus(template)
        query_params = self.current_request.query_params
        context = {}
        for query_param in query_params:
            context[query_param] = query_params.getlist(query_param)
        env = self.create_environment(loader=jinja2.DictLoader({'index.html': template}))
        template = env.get_template('index.html')
        render = template.render(**context)
        return {"render": render}



import traceback
from functools import partial, update_wrapper

LAMBDA_NAMESPACE = 'lambda'
COWORKS_NAMESPACE = 'coworks'


class XRayContextManager:

    def __init__(self, app, recorder):
        app.log.debug("initializing xray context manager")

        @app.before_first_activation
        def capture_routes(event, context):
            try:
                for path, route in app.routes.items():
                    for method, entry in route.items():
                        view_function = entry.view_function
                        cws_function = view_function.__cws_func__

                        def captured(_view_function, *args, **kwargs):
                            subsegment = recorder.current_subsegment()
                            if subsegment:
                                subsegment.put_metadata('event', event, LAMBDA_NAMESPACE)
                                subsegment.put_metadata('context', context, LAMBDA_NAMESPACE)
                                subsegment.put_annotation('service', app.name)
                                subsegment.put_metadata('query_params', app.current_request.query_params,
                                                        COWORKS_NAMESPACE)
                                if app.current_request.raw_body:
                                    subsegment.put_metadata('json_body', app.current_request.json_body,
                                                            COWORKS_NAMESPACE)
                            response = _view_function(*args, **kwargs)
                            if subsegment:
                                subsegment.put_metadata('response', response, COWORKS_NAMESPACE)
                            return response

                        def captured_entry(_cws_function, *args, **kwargs):
                            subsegment = recorder.current_subsegment()
                            if subsegment:
                                subsegment.put_metadata('args', args, COWORKS_NAMESPACE)
                                subsegment.put_metadata('kwargs', kwargs, COWORKS_NAMESPACE)
                            response = _cws_function(*args, **kwargs)
                            if subsegment:
                                subsegment.put_metadata('response', response, COWORKS_NAMESPACE)
                            return response

                        wrapped_fun = update_wrapper(partial(captured, view_function), view_function)
                        entry.view_function = recorder.capture(view_function.__name__)(wrapped_fun)

                        wrapped_fun = update_wrapper(partial(captured_entry, cws_function), cws_function)
                        entry.view_function.__cws_func__ = recorder.capture(cws_function.__name__)(wrapped_fun)
            except Exception:
                app.log.error("Cannot set xray context manager : are you using xray_recorder?")
                raise

        @app.handle_exception
        def capture_exception(event, context, e):
            try:
                app.log.error(f"Event: {event}")
                app.log.error(f"Context: {context}")
                app.log.error(f"Exception: {str(e)}")
                app.log.error(traceback.print_exc())
                subsegment = recorder.current_subsegment
                if subsegment:
                    subsegment.put_annotation('service', app.name)
                    subsegment.add_exception(e, traceback.extract_stack())
            finally:
                return {
                    'headers': {},
                    'multiValueHeaders': {},
                    'statusCode': 500,
                    'body': "Exception in microservice, see logs in XRay for more details",
                    'error': str(e)
                }



from copy import deepcopy
from dataclasses import dataclass, asdict
from logging import getLogger, WARNING

import anyconfig
import click
import sys
from typing import List, Tuple

from .command import CwsMultiCommands
from .error import CwsClientError
from ..config import DEFAULT_PROJECT_DIR, DEFAULT_WORKSPACE
from ..utils import import_attr, get_system_info
from ..version import __version__

PROJECT_CONFIG_VERSION = 2


@click.group()
@click.version_option(version=__version__, message=f'%(prog)s %(version)s, {get_system_info()}')
@click.option('-p', '--project-dir', default=DEFAULT_PROJECT_DIR,
              help=f"The project directory path (absolute or relative) [default to '{DEFAULT_PROJECT_DIR}'].")
@click.option('-c', '--config-file', help="Configuration file path [path from project dir].")
@click.option('-m', '--module', help="Filename of your microservice python source file.")
@click.option('-s', '--service', help="Microservice variable name in the source file.")
@click.option('-w', '--workspace', default=DEFAULT_WORKSPACE,
              help=f"Application stage [default to '{DEFAULT_WORKSPACE}'].")
@click.pass_context
def client(*args, **kwargs):
    ...


def invoke(ctx):
    """Invokes the command over the service or the declared services in project configuration file."""
    try:
        args = ctx.args
        protected_args = ctx.protected_args
        if not protected_args:
            sys.stderr.write(str("No command given.\n"))
            client.main(['--help'])
            sys.exit(1)
        command_name = protected_args[0]

        # get project options
        cws_options = CwsClientOptions(ctx.params)
        if not cws_options.services:
            sys.stderr.write(str("Nothing to execute as no service defined.\n"))
            sys.exit(1)
        project_dir = cws_options.project_dir
        workspace = cws_options.workspace

        # Iterates over the declared services in project configuration file
        commands_to_be_executed = CwsMultiCommands()
        for module, service in cws_options.services:
            ctx.args = list(args)
            ctx.protected_args = protected_args

            # Get command from the microservice description
            handler = cws_options.get_handler(module, service)
            handler.deferred_init(workspace)
            service_config = cws_options.get_service_config(module, service)
            command = service_config.get_command(command_name, handler)
            if not command:
                raise CwsClientError(f"Undefined command {command_name}.\n")
            command_options = service_config.get_command_options(command_name)

            # Get user defined options and convert them in right types
            client_options, _, cmd_opts = command.make_parser(ctx).parse_args(ctx.args)
            for opt_key, opt_value in client_options.items():
                cmd_opt = next(x for x in cmd_opts if x.name == opt_key)
                client_options[opt_key] = cmd_opt.type(opt_value)

            # Adds command and global options
            options = {**command_options, **client_options, '_from_cws': True}
            if options.get('help', False):
                print(command.get_help(ctx))
                return
            command.make_context(command.name, options)
            commands_to_be_executed.append(command, options)

        # Executes all commands
        for command_class, execution_list in commands_to_be_executed.items():
            command_class.multi_execute(project_dir, workspace, execution_list)
    except CwsClientError as client_err:
        sys.stderr.write(f"Error in command: {client_err.msg}\n")
        sys.exit(1)
    except Exception as e:
        sys.stderr.write(f"Error in command: {str(e)}\n")
        sys.exit(1)


client.invoke = invoke


@dataclass
class CwsClientOptions:
    """Client options defined from click command."""
    project_dir: str
    workspace: str
    module: str
    service: str
    config_file: str

    def __init__(self, params):
        self.project_dir = params.get('project_dir')
        self.workspace = params.get('workspace')
        self.module = params.get('module')
        self.service = params.get('service')
        self.config_file = params.get('config_file') or "project"
        self.project_config = ProjectConfig(self.project_dir, self.config_file)

    @property
    def services(self):
        """Returns the list of services defined from the client optons"""
        if self.service:
            return [(self.module, self.service)]
        return self.project_config.all_services(self.module)

    def get_handler(self, module, service):
        # Load microservice handler
        try:
            return import_attr(module, service, cwd=self.project_dir)
        except AttributeError as e:
            raise CwsClientError(f"Module '{module}' has no microservice {service} : {str(e)}\n")
        except ModuleNotFoundError as e:
            raise CwsClientError(f"The module '{module}' is not defined in {self.project_dir} : {str(e)}\n")
        except Exception as e:
            raise CwsClientError(f"Error {e} when loading module '{module}'\n")

    def get_service_config(self, module, service, workspace=None):
        """Returns the microserrvice's configuration."""
        workspace = workspace or self.workspace
        return ServiceConfig(self.project_config, module, service, workspace)


class ProjectConfig:
    """Class for the project configuration file."""

    def __init__(self, project_dir, file_name, file_suffix=".cws.yml"):
        from pathlib import Path
        self.project_dir = project_dir
        self.params = {}

        project_dir_path = Path(project_dir)
        self.project_file = project_dir_path / (file_name + file_suffix)
        project_secret_file = project_dir_path / (file_name + '.secret' + file_suffix)
        getLogger('anyconfig').setLevel(WARNING)
        self.params = anyconfig.multi_load([self.project_file, project_secret_file], ac_ignore_missing=True)

        if self.params.get('version') != PROJECT_CONFIG_VERSION:
            raise CwsClientError(f"Wrong project file version (should be {PROJECT_CONFIG_VERSION}).\n")

    def get_service_config(self, module, service, workspace):
        return ServiceConfig(self, module, service, workspace)

    def all_services(self, module: str = None) -> List[Tuple[str, str]]:
        """ Returns the list of (module, microservice) on which the command will be executed."""
        services = self.params.get('services', {})

        res = []
        for s in services:
            if 'module' not in s or 'services' not in s:
                raise CwsClientError(f"Services wrongly defined.\n")

            if module and s['module'] != module:
                continue

            if 'services' in s:
                _module = s['module']
                _services = s['services']
                if type(_services) is str:
                    res.append((_module, _services))
                else:
                    for service in _services:
                        res.append((_module, service))
        return res

    @property
    def all_commands(self):
        """ Returns the list of microservices on which the command will be executed."""
        return self.params.get('commands', {})

    @staticmethod
    def _get_workspace_options(options, workspace):
        """Returns the option values defined for the specific workspace or globally."""
        workspaces = options.pop('workspaces', {})
        workspace_options = {k: v for x in workspaces if x.pop('workspace', None) == workspace
                             for k, v in x.items()}
        return {**options, **workspace_options}

    def _get_service_options(self, services, service, workspace):
        """Returns the option values defined for the specific service and workspace or globally."""

        service_options = {}
        for s in services:
            if s.pop('service', None) == service:
                s.pop('module', None)
                service_options.update(self._get_workspace_options(s, workspace))
        return {**service_options}

    def get_module_options(self, options_list, module, service, workspace):
        """Returns the option values defined for the specific module, service and workspace or globally."""

        if type(options_list) is not list:
            options_list = [options_list]

        service_options = {}
        module_options = {}
        for options in options_list:
            if 'module' not in options or options.pop('module') == module:
                services = options.pop('services', {})
                module_options.update(self._get_workspace_options(options, workspace))
                service_options.update(self._get_service_options(services, service, workspace))

        return {**module_options, **service_options}


@dataclass
class ServiceConfig:
    project_config: ProjectConfig
    module: str
    service: str
    workspace: str

    @property
    def client_params(self):
        res = asdict(self)
        del res['project_config']
        res['project_dir'] = self.project_config.project_dir
        return res

    def get_command(self, cmd_name, ms):
        """Get the command associated to this microservice."""

        # Get command already added in handler
        for name in ms.commands:
            if name == cmd_name:
                return ms.commands[name]

        # Creates it from project class parameter if not already defined
        cmd_class = self._command_class(cmd_name)
        if cmd_class:
            cmd = cmd_class(ms, name=cmd_name)

            # Installs needed commands
            for needed in cmd.needed_commands:
                self.get_command(needed, ms)

            return cmd

    def _command_class(self, cmd_name):
        cmd_class_name = self.get_command_options(cmd_name).get('class')
        if cmd_class_name:
            splitted = cmd_class_name.split('.')
            return import_attr('.'.join(splitted[:-1]), splitted[-1], cwd=self.project_config.project_dir)

    def get_command_options(self, cmd_name):
        options = deepcopy(self.project_config.all_commands.get(cmd_name, {}))
        module_options = self.project_config.get_module_options(options, self.module, self.service, self.workspace)
        return {**self.client_params, **module_options}


def main():
    return client()


if __name__ == "__main__":
    main()



import contextlib
import logging
import os
import socket
import sys
from threading import Thread, Event

import click
from aws_xray_sdk.core import xray_recorder
from chalice.cli import run_local_server, reloader
from chalice.config import Config
from chalice.local import LocalDevServer

from .command import CwsCommand


class CwsRunner(CwsCommand):
    @property
    def options(self):
        return [
            *super().options,
            click.option('-h', '--host', default='127.0.0.1'),
            click.option('-p', '--port', default=8000, type=click.INT),
            click.option('--autoreload', is_flag=True, help='Reload server on source changes.'),
            click.option('--debug', is_flag=True, help='Print debug logs to stderr.')
        ]

    @classmethod
    def multi_execute(cls, project_dir, workspace, execution_list):
        """ Runs the microservice in a local Chalice emulator.
        """

        # chalice.cli package is not defined in deployment
        from .factory import CwsFactory

        for command, options in execution_list:
            os.environ['WORKSPACE'] = workspace
            command.app.config.load_environment_variables(project_dir)
            if command.app.entries is None:
                command.app.deferred_init(workspace)

            debug = options['debug']
            autoreload = options['autoreload']
            if debug:
                logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(message)s')

            if autoreload:
                if not options['_from_cws']:
                    sys.argv = [sys.executable, sys.argv[0]]

                class _ThreadedLocalServer(ThreadedLocalServer):

                    def __init__(self):
                        super().__init__()
                        self._app_object = self
                        self._config = Config()
                        self._host = options['host']
                        self._port = options['port']
                        self._server = LocalDevServer(command.app, self._config, self._host, self._port)

                rc = reloader.run_with_reloader(_ThreadedLocalServer, os.environ, project_dir)
                sys.exit(rc)
            else:
                factory = CwsFactory(command.app, project_dir, debug=debug)
                run_local_server(factory, options['host'], options['port'], workspace)

    def __init__(self, app=None, name='run'):
        super().__init__(app, name=name)
        xray_recorder.configure(context_missing="LOG_ERROR")


class ThreadedLocalServer(Thread):
    def __init__(self, *, port=None, host='localhost'):
        super().__init__()
        self._app_object = None
        self._config = None
        self._host = host
        self._port = port or self.unused_tcp_port()
        self._server = None
        self._server_ready = Event()

    def configure(self, app_object, config=None, **kwargs):
        self._app_object = app_object
        self._config = config

    def run(self):
        self._server = LocalDevServer(self._app_object, self._config, self._host, self._port, )
        self._server_ready.set()
        self._server.serve_forever()

    def serve_forever(self):
        self._server.serve_forever()

    def shutdown(self):
        if self._server is not None:
            self._server.server.shutdown()

    def make_call(self, method, path, timeout=0.5, **kwarg):
        self._server_ready.wait()
        return method(f'http://{self._host}:{self._port}{path}', timeout=timeout, **kwarg)

    @classmethod
    def unused_tcp_port(cls):
        with contextlib.closing(socket.socket()) as sock:
            sock.bind(('localhost', 0))
            return sock.getsockname()[1]



import json

from chalice import BadRequestError, NotFoundError

from ..coworks import TechMicroService


class S3MicroService(TechMicroService):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__s3_client__ = None

    @property
    def s3_client(self):
        if self.__s3_client__ is None:
            self.__s3_client__ = self.boto3_session.client('s3')
        return self.__s3_client__

    def get_bucket(self):
        """Return the list buckets defined."""
        return json.dumps(self.s3_client.list_buckets(), indent=4, sort_keys=True, default=str)

    def get_bucket_(self, bucket, key=None):
        """Returns an object in the bucket from the key or the list of objects contained in the bucket."""
        if key:
            uploaded_object = self.s3_client.get_object(bucket=bucket, key=key)
            if not uploaded_object:
                raise NotFoundError(f"No key '{key}' in bucket '{bucket}'")
            return uploaded_object["Name"]
        else:
            b = self.s3_client.list_objects(bucket=bucket)
            if b is None:
                raise NotFoundError(f"No  bucket '{bucket}'")
            return {'Name': bucket}

    def put_bucket(self, bucket):
        kwargs = {}
        found = self.s3_client.list_objects(bucket=bucket)
        if found is not None:
            raise BadRequestError(f"Bucket '{bucket}' already exists.")
        if self.region_name != 'us-east-1':
            kwargs['createBucketConfiguration'] = {'LocationConstraint': 'EU'}
        self.s3_client.create_bucket(bucket=bucket, **kwargs)

    def delete_bucket(self, bucket, key=None):
        if key is None:
            self.s3_client.delete_bucket(bucket=bucket)
        else:
            self.s3_client.delete_object(bucket=bucket, key=key)

    def get_content(self, bucket, key=None):
        if key:
            uploaded_object = self.s3_client.get_object(bucket=bucket, key=key)
            if not uploaded_object:
                return NotFoundError(f"No key '{key}' in bucket '{bucket}'")
            return uploaded_object['Body'].read().decode('utf-8')
        else:
            raise BadRequestError("Key is missing.")

    def put_content(self, bucket, key=None, body=b''):
        if key:
            self.s3_client.put_object(bucket=bucket, key=key, body=body)
        else:
            raise BadRequestError("Key is missing.")



import pprint
import sys
from abc import ABC, abstractmethod
from collections import defaultdict

import click

from .error import CwsCommandError
from ..coworks import TechMicroService


class CwsClientCommandOptions:

    def __init__(self, client_options, execution_context):
        self.client_options = client_options
        self.execution_context = execution_context

    def get(self, option, default_value=None):
        return self.client_options.get(option, default_value)

    def pop(self, option, default_value):
        """Removes from client option and all command options."""
        value = self.client_options.pop(option, default_value)
        for execution_params in self.execution_context.values():
            for command, command_options in execution_params:
                command_options.pop(option, None)
        return value

    def __repr__(self):
        str = "client options:"
        str += pprint.saferepr(self.client_options) + '\n'
        for execution_params in self.execution_context.values():
            for command, command_options in execution_params:
                str += f"command options for {command.app.name} : "
                str += pprint.saferepr(command_options)
                str += '\n'
        return str


class CwsMultiCommands:
    def __init__(self):
        self.execution_context = defaultdict(list)

    def append(self, command, command_options):
        self.execution_context[type(command)].append((command, command_options))

    def items(self):
        for command_class, execution_params in self.execution_context.items():
            yield command_class, execution_params


class CwsCommand(click.Command, ABC):

    @classmethod
    @abstractmethod
    def multi_execute(cls, project_dir, workspace, execution_list):
        """ Main command execution function.
         :param project_dir: Project directory where projet configuration file can be found.
         :param workspace: Environment workspace.
         :param execution_list: List of (command, command options from projet configuraion file).

         Abstract method which must be redefined in any subclass. The content should be written in self.output.
         """

    def __init__(self, app: TechMicroService = None, *, name):
        super().__init__(name, callback=self.__execute)

        # Trace interfaces.
        self.output = sys.stdout
        self.error = sys.stderr

        # A list of functions that will be called before or after the command executioon.
        self.before_funcs = []
        self.after_funcs = []

        if app is not None:
            self.app = app
            self.init_app(app)

        for opt in self.options:
            opt(self)

    def init_app(self, app):
        app.commands[self.name] = self
        for cmd in self.needed_commands:
            if cmd not in app.commands:
                raise CwsCommandError(f"Undefined command {cmd} needed.")

    @property
    def needed_commands(self):
        return []

    @property
    def options(self):
        return []

    def execute(self, *, project_dir, module, service, workspace, output=None, error=None, **options):
        """ Called when the command is called.
        :param output: output stream.
        :param error: error stream.
        :param options: command options.
        :return: None
        """
        self.app.deferred_init(workspace)

        if output is not None:
            self.output = open(output, 'w+') if type(output) is str else output
        if error is not None:
            self.error = open(error, 'w+') if type(error) is str else error

        try:
            for func in self.before_funcs:
                func(options)

            ctx = self.make_context(self.name, options)
            ctx_options = {**options, 'output': output, 'error': error}
            ctx_options.setdefault('_from_cws', False)
            ctx.params.update(project_dir=project_dir, module=module, service=service, workspace=workspace,
                              **ctx_options)

            if ctx_options.get('help', False):
                print(self.get_help(ctx))
                return

            self.invoke(ctx)

            for func in self.after_funcs:
                func(options)
        except click.exceptions.Exit as e:
            if e.exit_code:
                raise CwsCommandError("Command exits with error")
            return
        except CwsCommandError:
            raise
        except Exception as e:
            raise CwsCommandError(str(e))

    def before_execute(self, f):
        """Registers a function to be run before the command execution.
        :param f: function called before the command execution
        :return: None

        May be used as a decorator.

        The function will be called without any arguments and its return value is ignored.
        """

        self.before_funcs.append(f)
        return f

    def after_execute(self, f):
        """Registers a function to be run after the command execution.
        :param f: function called after the command execution
        :return: None

        May be used as a decorator.

        The function will be called without any arguments and its return value is ignored.
        """

        self.after_funcs.append(f)
        return f

    def parse_args(self, ctx, args):
        for param in self.get_params(ctx):
            if param.name not in args:
                if param.required:
                    raise CwsCommandError(f"missing parameter: {param.name}")
                args[param.name] = param.get_default(ctx)
        ctx.args = args
        return args

    def __execute(self, **options):
        project_dir = options['project_dir']
        workspace = options['workspace']
        self.multi_execute(project_dir=project_dir, workspace=workspace, execution_list=[(self, options)])
